{% extends "base.html" %}

{% block title %}Runs{% endblock %}

{% block content %}
<table>
    <thead>
        <tr>
            <th>Date</th>
            <th>Run ID</th>
        </tr>
    </thead>
    <tbody>
        {% for item in runs %}
        <tr>
            <td>{{ item.date }}</td>
            <td><a href="{{ item.run_id }}">{{ item.run_id }}</a></td>
        </tr>
        {% endfor %}
    </tbody>
</table>
{% endblock %}
