<!DOCTYPE html>
<html lang="en">

<head>
    {% block head %}
    <link rel="stylesheet" href="/static/style.css" />
    <title>{% block title required %}{% endblock %} - GlobalFit Monitoring</title>
    {% endblock %}
</head>

<body>
    <div id="content">{% block content %}{% endblock %}</div>
    <div id="footer">
        {% block footer %}
        {% endblock %}
    </div>
</body>

</html>
