{% extends "base.html" %}

{% block title %}Run details{% endblock %}

{% block content %}
<table>
    <thead>
        <tr>
            <th>Group</th>
            <th>Modules</th>
        </tr>
    </thead>
    <tbody>
        {% for item in groups %}
        <tr>
            <td>{{ item.group_id }}</td>
            <td>
                {% for module_id in item.modules %}
                <a href="groups/{{item.group_id}}/modules/{{ module_id }}">{{ module_id }}</a>
                {% endfor %}
            </td>
        </tr>
        {% endfor %}
    </tbody>
</table>
{% endblock %}
