{% extends "base.html" %}

{% block title %}Home{% endblock %}


{% block content %}
<ul>
    {% for item in navigation %}
    <li><a href="{{ item.href }}">{{ item.caption }}</a></li>
    {% endfor %}
</ul>
{% endblock %}
